from numpy import arange, amin, amax, zeros, logical_and, logical_or, logical_xor, argmax
from pandas import read_table, DataFrame, date_range, concat
from sys import exit
from io import StringIO
from subprocess import check_output, CalledProcessError, STDOUT, Popen, PIPE
from tools import filter, unix, arraystr, start_timer, stop_timer
import conf
import os
import sys
import resource
'''
  -A  Use convex-hull approximation.
  -C  Disable most memory reduction techniques.
  -E  Find best solution (same as -t3, but wihtout printing the trace).
  -H n
      Set hash table size for bit state hashing to 2**n
      (default = 27)
  -n <0|1|2|3|4>
      Select extrapolation operator.
        0: Automatic
        1: No extrapolation (use with care)
        2: Difference extrapolation
        3: Location based extrapolation
        4: Lower/upper extrapolation
  -o <0|1|2|3|4|5>
      Select search order.
        0: Breadth first (default)
        1: Depth first
        2: Random depth first
        3: Optimal first
        4: Random optimal depth first
        5: Smallest heur first
  -S <0|1|2>
      Optimize space consumption (0 = none, 1 = default, 2 = most)
  -T  Reuse state space when several properties are examined.
  -Z  Use bit-state hashing.

Options for trace generation:
  -f prefix
      Write symbolic traces to files 'prefix-n.xtr' rather than to stderr.
  -t <0|1|2>
      Generate diagnostic information on stderr.
        0: Some trace
        1: Shortest trace (disables reuse)
        2: Fastest trace (disables reuse)
        3: Best trace (disables reuse, enables -o4)
  -y  Display traces symbolically (post-stable).
  -Y  Display traces symbolically (pre- and post-stable).

Miscellaneous options:
  -h  Shows this help screen.
  -q  Do not display the copyright message.
  -r seed
      Set seed for random number generator (default is current time).
  -s  Do not display the progress indicator.
  -u  Show summary after verification (incorrect for liveness properties).
  -v  Show version number.

'''

# Run Cots-Optimal Reachability Analysis on model with query.
# model -- path to model file
# query -- path to query file

def cora():

	try:
		now = start_timer()

		p = Popen([
			conf.param.binaries + 'verifyta',
			'-t 3',
			'-T',
			'-u',
			'-S 2',
			'-n 3',
			conf.param.model_file,
			conf.param.query_file,
		], stderr=PIPE)
		out,err = p.communicate()
		conf.times['Cora'] = stop_timer(now)
		#clean up
		if conf.param.temporary:
			os.unlink(conf.param.model_file)
			os.unlink(conf.param.query_file)

		return out,err

	except OSError as e:
		print("OSError > " + str(e.errno))
		print("OSError > " + str(e.strerror))
		print("OSError > " + str(e.filename))
		exit()
	except:
		exit("Error > " + str(sys.exc_info()[1]))

# This function parses UPPAAL Cora output
# (generated by ./verifyta) into Pandas DataFrames
# vars for the variable valuations and states for
# the locations the automata are in.

def parse(fname):
	if isinstance(fname, str):
		f = open(fname, 'r')
	elif isinstance(fname, StringIO):
		f = fname
	else:
		exit('fname must either be a path to a file or StingIO')

	var_lst, state_lst, trans_lst = [], [], []
	for block in f.read().split('\n\n'):
		if 'State:' in block:
			parts = block.split('\n')
			state_lst.append(parts[1][2:-2].split(' '))
			var_lst.append(parts[2].split(' '))
		## PARSING TRANSITIONS MIGHT NOT BE NECESSARY
	f.close()
	name	= lambda sep,x: x.split(sep)[0]
	value	= lambda sep,x: x.split(sep)[1]

	vars = DataFrame(
		data = var_lst,
		columns = [name('=',x) for x in var_lst[0]],
	).applymap(lambda x:int(value('=', x)))

	states = DataFrame(
		data = state_lst,
		columns = [name('.',x) for x in state_lst[0]],
	).applymap(lambda x:value('.', x))

	return concat([vars, states], axis=1)

def trans_state(state):
	# non-array variables
	mask = state.index.str.endswith(']')
	trans = dict(state[logical_not(mask)])
	# array variables
	while state[mask].shape[0]:
		name = str(state[mask].index[0]).split('[')[0]
		tmp = logical_and(mask, state.index.str.startswith(name+'['))
		trans[name] = arraystr(state[tmp], ',')
		mask = logical_xor(mask,tmp)
	return trans

# Save number of scheduled jobs and some diagnostic values in a dictionary
# Returns the dictionary
def aggregate(trace,values):
	dict = {
		'Uppaal Cost' : int(trace.iloc[-1]['cost']),
		'Minimal SoC' : int(amin(trace['soc'])),
		'Maximal SoC' : int(amax(trace['soc'])),
		'Remain. SoC' : int(trace['soc'].iloc[-1])
	}
	# Add Accesses
	for val in [v for v in values if v['id'] != 'Sun']:
		dict[val['id']] = int(sum([trace.iloc[-1]['nee[{0}]'.format(i)] for i in val['idx']]))

	return dict

# plots the SoC curve from the UPPAAL output
def plotSoC(ax, data, xticks):
	import matplotlib.pyplot as plt
	soc, = ax.plot(data['gc'],data['soc']/conf.param.capacity)
	ax.set(
		xlim = [unix(conf.param.start), unix(conf.param.end)],
		ylim = [conf.param.cora_safe_mode, 1],
		yticks = arange(0,1.1,0.1),
		xticks = unix(xticks),
		xticklabels = []
	)
	return soc

# plots the Load step function from the UPPAAL output
def plotLoad(ax, data, xticks, color='purple',lw=3):
	import matplotlib.pyplot as plt
	load, = ax.step(data['gc'], data['B.load'], c = color, lw = lw)
	ax.set(
		xlim = [unix(conf.param.start), unix(conf.param.end)],
		ylim = [-3200, amax(data['B.load'])],
		yticks = [-3200, 0, amax(data['B.load'])],
		#xticks = data['gc']
		xticks = unix(xticks),
		xticklabels = xticks.strftime("%d@%H:%M")
	)
	return load

